#!/usr/bin/env bash
#shellcheck disable=SC2034,SC2050

########################################################################################################################
#
# This job is the wrapper for cmor jobs. It is meant to setup the environment, create the cmorization scripts, and 
# execute them using Parsl
#
# VARIABLE INDEX:
#
# Mind that these variables get substituted everywhere, including inside of the origina cmor script.
#
# internal autosubmit:
#
# CURRENT_ROOTDIR: path to files of the experiment in remote
# EXPID: experiment id of Autosubmit
# SDATE: start date of this workflow
# CHUNK: chunk of this workflow
# JOBTYPE: name without expid, stardate, and chunk of the job (as it appears 
#          in the section)
#
# in expdef_cxxx.conf (if you use TODO INSERT HASH) data you should not change this:
#
# ECEARTH_OUTPUT_SDATE: original data startdate. It should match SDATE
# ECEARTH_OUTPUT_MEMBER: original data member (fc0, fc1). It shoud match MEMBER
# ECEARTH_OUTPUT_MEMBERS: list of members of the original data. It should match MEMBERS
# ECEARTH_OUTPUT_MODEL_RES: resolution of the original data
# ECEARTH_OUTPUT_VERSION: model version that produced the original data
# ECEARTH_OUTPUT_BSC_OUTCLASS: cmorization configuration from bsc of the original data
# ECEARTH_OUTPUT_CMOR_ACTIVITY_ID: identification of the cmor execise 
# ECEARTH_OUTPUT_CMOR_EXP: experiment of the cmor (pre-industrial control, piControl, etc)
# ECEARTH_OUTPUT_CMOR_MODEL_ID: id of the model to produce original data
# ECEARTH_OUTPUT_LPJG: boolean to indicate if lpjg is enabled
# ECEARTH_OUTPUT_CHUNK: original data chunk
# ECEARTH_OUTPUT_CHUNK_START_DATE: original data Chunk Start date
# ECEARTH_OUTPUT_CHUNK_END_DATE: original data chunk's end date
# ECEARTH_OUTPUT_CHUNK_LAST: lst chunk to be cmorized
# ECEARTH_OUTPUT_CHUNKSIZE: lenght in months of the chunk (12 months, 6, etc)
# ECEARTH_OUTPUT_NFRP: IFS configuration
# ECEARTH_OUTPUT_NEMO_RESOLUTION: nemo grid indentifier of the original data
# ECEARTH_OUTPUT_IFS_RESOLUTION: ifs grid identifier of the original data
# ECEARTH_OUTPUT_PROJ_TYPE: original data project type
# ECEARTH_OUTPUT_PRODUCTION_EXP: boolean indicating if the original data is production experiment
# ECEARTH_OUTPUT_CMORIZATION: boolean that enables/disables cmorization
# ECEARTH_OUTPUT_CMOR_EXP_CUSTOM: custom flag for cmorization
# ECEARTH_OUTPUT_CMOR_ADD_STARTDATE: custom startdata for cmorization
# ECEARTH_OUTPUT_TEMPLATE_NAME: original data template name
#
# in autosubmit_cxxx.conf: 
#
# CMOR_WRAPPED_COMPONENTS = atm oce
#
########################################################################################################################

ROOTDIR=%CURRENT_ROOTDIR%
LOGDIR=$ROOTDIR/LOG_%EXPID%

#
# Manu: we will use the name of the job, which should 
#
number_of_members=$(echo %TASKTYPE% | cut -d "_" -f 2 | sed -e 's/m//')
number_of_chunks=$(echo %TASKTYPE% | cut -d "_" -f 3 | sed -e 's/c//')

# loop over every member to be executed
for THIS_MEMBER_NUMBER in $(seq $number_of_members); do

  THIS_MEMBER=$(printf "%02d\n" $THIS_MEMBER_NUMBER)

  # loop over every chunk to be executed
  for THIS_CHUNK_NUMBER in $(seq $number_of_chunks); do 

  THIS_CHUNK=$(printf "%03d\n" $THIS_CHUNK_NUMBER)
  
    # loop over every component to be cmorized
    for component in %CMOR_WRAPPED_COMPONENTS%; do 

      # cmorization output path
      cmor_outpath=cmor_outputs/cmor_${component}_${THIS_MEMBER}_${THIS_CHUNK}
    
      script_name=%EXPID%_%SDATE%_${THIS_MEMBER}_${THIS_CHUNK}_CMOR$(echo $component | tr '[:lower:]' '[:upper:]').cmd 

      # output the full script as is using HEREDOCS so that autosubmit does the 
      # variable change but does not evaluate anything else.
      # we add quotes to avoid command substitution :)
      cat > $script_name << "EOF" 
#!/usr/bin/env bash
#shellcheck disable=SC2034,SC2050

########################################################################################################################
# this job deals with the standardization of the output (cmorization) depending on the outclass used
# intended to be executed from the remote machine, can run in Nord3v2 too but at cost of some efficiency
########################################################################################################################

#
# Architecture
#
CURRENT_ARCH=%CURRENT_ARCH%
EXPID=%EXPID%
GROUP=%CURRENT_PROJ%
ROOTDIR=%CURRENT_ROOTDIR%
PROJDEST=%PROJECT_DESTINATION%
PROJDIR=$ROOTDIR/$PROJDEST
BSC_OUTCLASS=%ECEARTH_OUTPUT_BSC_OUTCLASS%
CMIP6_OUTCLASS=
JOBNAME=%JOBNAME%
CMOR_REALIZATION_INDEX=""
[[ "%ECEARTH_OUTPUT_LPJG%" == TRUE ]] && LPJG=TRUE || LPJG=FALSE
TABLES_DIR=${PROJDIR}/ece2cmor/ece2cmor3/resources/tables
CMOR_INITIALIZATION_INDEX="1"

export HPCPROJ

#
# General Paths and Conf.
#
MODEL=ecearth
MODEL_RES=%ECEARTH_OUTPUT_MODEL_RES%
VERSION=%ECEARTH_OUTPUT_VERSION%
START_date=%ECEARTH_OUTPUT_SDATE%
MEMBER=%ECEARTH_OUTPUT_MEMBER%
MEMBERS_LIST=%ECEARTH_OUTPUT_MEMBERS%
SCRATCH_DIR=%SCRATCH_DIR%
CMOR_ACTIVITY_ID=%ECEARTH_OUTPUT_CMOR_ACTIVITY_ID%
CMOR_EXP=%ECEARTH_OUTPUT_CMOR_EXP%
CMOR_MODEL_ID=%ECEARTH_OUTPUT_CMOR_MODEL_ID%
TEMPLATE_NAME=%ECEARTH_OUTPUT_TEMPLATE_NAME%
TASKTYPE=%TASKTYPE%
#
# Chunk Management
#
CHUNK=%ECEARTH_OUTPUT_CHUNK%
Chunk_start_date=%ECEARTH_OUTPUT_CHUNK_START_DATE%
Chunk_end_date=%ECEARTH_OUTPUT_CHUNK_END_DATE%
Chunk_last=%ECEARTH_OUTPUT_CHUNK_LAST%
# In months
CHUNKSIZE=%ECEARTH_OUTPUT_CHUNKSIZE%

RUN_dir=${ROOTDIR}/${START_date}/${MEMBER}/runtime
output_dir_atm=${RUN_dir}/output/ifs/$(printf "%03d\n" $CHUNK)
output_dir_oce=${RUN_dir}/output/nemo/$(printf "%03d\n" $CHUNK)

#
# Cmorization
#
NFRP=%ECEARTH_OUTPUT_NFRP%
NEMO_resolution=%ECEARTH_OUTPUT_NEMO_RESOLUTION%
IFS_resolution=%ECEARTH_OUTPUT_IFS_RESOLUTION%
PROJ_TYPE=%ECEARTH_OUTPUT_PROJ_TYPE%
PRODUCTION_EXP=%ECEARTH_OUTPUT_PRODUCTION_EXP%
CMORIZATION=%ECEARTH_OUTPUT_CMORIZATION%
CMOR_EXP_CUSTOM=%ECEARTH_OUTPUT_CMOR_EXP_CUSTOM%
CMOR_ADD_STARTDATE=%ECEARTH_OUTPUT_CMOR_ADD_STARTDATE%

#####################################################################################################################
# Globals: PROJDIR, CURRENT_ARCH, CMIP6_OUTCLASS, BSC_OUTCLASS, MODEL_RES, NEMO_resolution, IFS_resolution, MEMBER,
#         Chunk_start_date, CMOR_EXP_CUSTOM, PATH_FLAGS, CMOR_ACTIVITY_ID, CMOR_MODEL_ID
# Arguments:
#
# Returns:
#   None
# Purpose: Prepare que environment for running the cmorization process, this method is common to all type of models.
#          it creates the neccesary folders and defines the list of vars to be used depending of the defined OUTCLASS
#          type
#
#####################################################################################################################
function setenvcmor() {

  set +ue

  #shellcheck disable=SC1090
  . ${PROJDIR}/platforms/${CURRENT_ARCH}/configure.sh
  load_modules_ece2cmor
  define_paths_ece2cmor

  # If you want to use your own tables and/or your modified version of ece2cmor
  # and are running CMOR jobs on the nord3v2 platform
  # uncomment the last lines of the define_paths_ece2cmor() function in platforms/nord3v2/configure.sh

  . ${PROJDIR}/plugins/cmorization.sh

  # set varlist
  if [[ -z ${CMIP6_OUTCLASS-} && -z ${BSC_OUTCLASS-} ]]; then
    echo "Classic outclass can't be CMORIZED"
    exit 1
  fi

  if [[ -n "${CMIP6_OUTCLASS}" ]] && [[ -n "${BSC_OUTCLASS}" ]]; then
    echo "only one OUTCLASS should be set"
    exit 1
  fi

  if [[ -n "${BSC_OUTCLASS-}" ]] && [[ -z "${CMIP6_OUTCLASS-}" ]]; then
    REPLACEMENT_TABLES=$(find ${PROJDIR}/outclass/${BSC_OUTCLASS} -name "CMIP6*.json" -type f)
    if [[ -z ${REPLACEMENT_TABLES} ]]; then
      echo "No CMIP6 tables have been replaced."
    else
      cp ${PROJDIR}/outclass/${BSC_OUTCLASS}/CMIP6*json ${PROJDIR}/ece2cmor/ece2cmor3/resources/tables/.
    fi
    EXCEL_FILE=$(echo ${PROJDIR}/outclass/${BSC_OUTCLASS}/*varlist*.json)
  fi

  if [[ -n "${CMIP6_OUTCLASS}" ]] && [[ -z "${BSC_OUTCLASS}" ]]; then
    if [[ ${CMIP6_OUTCLASS} == "OMIP/cmip6-experiment-OMIP-omip1" ]] && [[ ${CMOR_MODEL_ID} == "EC-EARTH-AOGCM" ]]; then
      EXCEL_FILE=${ROOTDIR}/ctrl/output-control-files/cmip6/${CMIP6_OUTCLASS}/cmip6-data-request-varlist-OMIP-omip1-EC-EARTH-CC.json
    elif [[ ${CMIP6_OUTCLASS} == "ScenarioMIP/EC-EARTH-CC/cmip6-experiment-ScenarioMIP-ssp245" ]] && [[ ${CMOR_MODEL_ID} == "EC-EARTH-AOGCM" ]]; then
      EXCEL_FILE=${ROOTDIR}/ctrl/output-control-files/cmip6/${CMIP6_OUTCLASS}/cmip6-data-request-varlist-ScenarioMIP-ssp245-EC-EARTH-CC.json
    else
      EXCEL_FILE=$(echo ${ROOTDIR}/ctrl/output-control-files/cmip6/${CMIP6_OUTCLASS}/*-varlist-*-*-*${CMOR_MODEL_ID}.json)
    fi
  fi
  ##

  if [[ ${MODEL_RES} == "LR" ]]; then
    MODEL_NAME="EC-Earth3"
  elif [[ ${MODEL_RES} == "HR" ]]; then
    MODEL_NAME="EC-Earth3-HR"
  elif [[ ${MODEL_RES} == "VHR" ]]; then
    MODEL_NAME="EC-Earth3-VHR"
  else
    echo "Wrong MODEL_RES" $MODEL_RES "is not in LR or HR or VHR"
    exit
  fi

  grid=$(echo $NEMO_resolution | cut -f1 -d "L")$(echo $IFS_resolution | cut -f1 -d "L")
  member_index=$((10#$(echo ${MEMBER} | cut -c3-) + 1))
  cmor_realization_indexing

  startyear=$(echo $Chunk_start_date | cut -c1-4)
  startmonth=$(echo $Chunk_start_date | cut -c5-6)
  [[ ! -f $ROOTDIR/LOG_$EXPID/${EXPID}_CMOR_DATASETVERSION ]] && echo "v$(date +%Y%m%d)" >$ROOTDIR/LOG_${EXPID}/${EXPID}_CMOR_DATASETVERSION
  #shellcheck disable=SC2155
  export cmordatasetversion=$(cat $ROOTDIR/LOG_${EXPID}/${EXPID}_CMOR_DATASETVERSION)
  mkdir -p ${RUN_dir}/cmor_outputs
  if [[ ${CMOR_EXP_CUSTOM} == "TRUE" ]]; then
    if ! grep -q $CMOR_EXP ${PROJDIR}/ece2cmor/ece2cmor3/resources/tables/CMIP6_CV.json; then
      python ${PROJDIR}/plugins/cmor_pythonsetup.py --projdir $PROJDIR --template_name $TEMPLATE_NAME --expid $EXPID --cmor_exp $CMOR_EXP
      if [[ $CMOR_ADD_STARTDATE == TRUE ]] || { ! grep -q $START_date ${PROJDIR}/ece2cmor/ece2cmor3/resources/tables/CMIP6_CV.json; }; then
        python ${PROJDIR}/plugins/cmor_pythonsetup.py --projdir $PROJDIR --template_name $TEMPLATE_NAME --expid $EXPID --cmor_exp $CMOR_EXP --cmor_add_startdate $CMOR_ADD_STARTDATE --startdate $START_date
      fi
    fi
  fi

  #TODO override defaults this should come from proj.conf
  #activity_id=CMIP
  activity_id=$CMOR_ACTIVITY_ID
  model_id=$CMOR_MODEL_ID

  mkdir -p ${RUN_dir}/cmor_outputs
  # path for cmor jobs flags/signaling
  export PATH_FLAGS=${RUN_dir}/flags
  if [[ ! -d ${PATH_FLAGS} ]]; then
    mkdir -p ${PATH_FLAGS}
  fi

  set -ue

}

#####################################################################################################################
# Globals: CMOR_ACTIVITY_ID, RUN_dir, CMOR_EXP, MEMBER, CMOR_EXP_CUSTOM, BSC_OUTCLASS, CMOR_ADD_STARTDATE, START_date
# Arguments:
#   None
# Returns:
#   None
# Purpose: based on the metadadata ( that depends of the type of cmorization to be run ) the cmor templates are generated
#
#####################################################################################################################
function cmor_gen_templates() {
  . ${PROJDIR}/plugins/cmorization.sh
  create_sub_experiment_id
  # cd into the metadata dir to avoid clashes with other cmor jobs
EOF
      # this is horrible, but we need variable substitution here! 
      cat >> $script_name << EOF 
  metadata_dir=\${RUN_dir}/../tmp/cmor/metadata-\${CMOR_ACTIVITY_ID}-\${component}-${THIS_MEMBER}-${THIS_CHUNK}-\${startyear}\${startmonth}
EOF
      cat >> $script_name << "EOF" 
  mkdir -p "${metadata_dir}"
  cd "${metadata_dir}"
  ${PROJDIR}/ece2cmor/ece2cmor3/scripts/modify-metadata-template.sh ${activity_id} ${CMOR_EXP} ${model_id} ${PROJDIR}/ece2cmor/ece2cmor3/resources/metadata-templates/cmip6-CMIP-piControl-metadata-template.json
  metadata_template_oce="${metadata_dir}/metadata*-${activity_id}-${CMOR_EXP}-${model_id}-nemo-template.json"
  metadata_template_atm="${metadata_dir}/metadata*-${activity_id}-${CMOR_EXP}-${model_id}-ifs-template.json"
  metadata_template_lpjg="${metadata_dir}/metadata*-${activity_id}-${CMOR_EXP}-${model_id}-lpjg-template.json"
  metadata_template_tm5="${metadata_dir}/metadata*-${activity_id}-${CMOR_EXP}-${model_id}-tm5-template.json"

  for f in "${metadata_dir}/"*"${CMOR_EXP}"*-*-template.json; do
    sed -i 's/"realization_index":            "1",/"realization_index":            "'"${realization_index:?}"'",/g' "$f"
EOF
      cat >> $script_name << EOF 
    sed -i '10 a\    "outpath":                      "'"\${RUN_dir}"'/$cmor_outpath",' "\$f"
EOF
      cat >> $script_name << "EOF" 
    sed -i '11 a\    "output_path_template":                      "<activity_id><institution_id><source_id><experiment_id><variant_label><table><variable_id><grid_label><'"$cmordatasetversion"'>",' "$f"
    sed -i '15 a\    "variant_info":                 "forcing: Nat.Ant. Member generated from autosubmit member '${MEMBER}'",' "$f"
    sed -i 's/"initialization_index":         "1",/"initialization_index":         "'${CMOR_INITIALIZATION_INDEX}'",/g' "$f"
    sed -i 's/"parent_activity_id":           "",/"parent_activity_id":         "CMIP",/g' "$f"
    sed -i 's/"parent_experiment_id":         "",/"parent_experiment_id":         "None",/g' "$f"
    if [[ -n $sub_experiment_id ]]; then
      sed -i 's/"sub_experiment_id":            "none",/"sub_experiment_id":            "'"${sub_experiment_id}"'",/g' "$f"
    fi
  cat "$f"
  done
}

#####################################################################################################################
# Globals: TASKTYPE, MEMBER, CHUNK, EXPID, EXCEL_FILE, RUN_dir, PROJDIR, PATH_FLAGS,
# Arguments:
#
# Returns:
#   None
# Purpose: Main cmorization function, common to most models
#
#####################################################################################################################
function cmor() {
  skip_alevel_vars=""
  if [[ ${component} == "oce" ]]; then
    cmor_model="nemo"
  elif [[ ${component} == "atm" ]]; then
    cmor_model="ifs"
    if [[ "$CMIP6_OUTCLASS" != *"pextra"* ]]; then
      skip_alevel_vars="--skip_alevel_vars"
    fi
  else
    cmor_model=${component}
  fi
  output_dir=output_dir_${component}
  metadata_template=metadata_template_${component}

  cmor_gen_templates
  cd "${metadata_dir}"
  find . -type f \! -name "*${cmor_model}*" -delete

EOF
      # horrible, but we need variable substitution here 
      cat >> $script_name << EOF 
  tmpdir=\${RUN_dir}/../tmp/cmor/tmp-\${component}-${THIS_MEMBER}-${THIS_CHUNK}-${THIS_MEMBER}-\${startyear}\${startmonth}
EOF
      # horrible, but we need variable substitution here 
      cat >> $script_name << "EOF" 
  rm -rf "$tmpdir"
  mkdir -p "$tmpdir"
EOF
      # horrible, but we need variable substitution here 
      cat >> $script_name << EOF 
  rm -rf \$RUN_dir/$cmor_outpath 
  mkdir -p \$RUN_dir/$cmor_outpath
EOF

      cat >> $script_name << "EOF" 
  cd ${PROJDIR}/ece2cmor/ece2cmor3
  #shellcheck disable=SC2086
  ./ece2cmor.py --${cmor_model} --exp ${EXPID} --meta ${!metadata_template} --tmpdir "${tmpdir}" --tabledir ${TABLES_DIR}  --varlist "${EXCEL_FILE}" "${!output_dir}" $skip_alevel_vars
  rm -rf "${tmpdir}"
  #fx management

  if [ ${CMOR_MODEL_ID} = "EC-EARTH-AOGCM" ]; then
    dataset="EC-Earth3"
  else
    dataset=${CMOR_MODEL_ID//EC-EARTH/EC-Earth3}
    # eg: EC-EARTH-HR -> EC-Earth3-HR, EC-EARTH-CC -> EC-Earth3-CC , EC-EARTH-AerChem -> EC-Earth3-AerChem, EC-EARTH-Veg-LR -> EC-EARTH-Veg-LR
  fi
#(always) overwrite sftof fx file due to ece2cmor issue https://github.com/EC-Earth/ece2cmor3/issues/483
  cmordatasetversion=$(cat $ROOTDIR/LOG_${EXPID}/${EXPID}_CMOR_DATASETVERSION)
  if [[ $component == "oce" ]]; then
    SFTOF_CMOR_FILENAME=sftof_Ofx_${dataset}_${CMOR_EXP}_r${realization_index:?}i${CMOR_INITIALIZATION_INDEX}p1f1_gn.nc
EOF
      cat >> $script_name << EOF 
    CMOR_OFX_DIR=\$RUN_dir/$cmor_outpath/\${CMOR_ACTIVITY_ID}/EC-Earth-Consortium/\${dataset}/\${CMOR_EXP}/r\${realization_index:?}i\${CMOR_INITIALIZATION_INDEX}p1f1/Ofx/sftof/gn/\$cmordatasetversion
EOF
      cat >> $script_name << "EOF" 
    mkdir -p "$CMOR_OFX_DIR"
    cp ${PROJDIR}/fx_files/${NEMO_resolution}/sftof_Ofx_EC-Earth3-CC_piControl_r0i0p0f0_gr.nc "${CMOR_OFX_DIR}/$SFTOF_CMOR_FILENAME"
  fi
  if [[ $(printf "%03d\n" ${CHUNK}) == '001' ]]; then
  #if 1st chunk and lsm, add fx areacella from the templates directory
    if [[ $component == "lpjg" ]] && [[ -n $IFS_resolution ]] ; then
      if [[ -n $sub_experiment_id ]]; then
        AREACELLA_CMOR_FILENAME=areacella_fx_${dataset}_${CMOR_EXP}_${sub_experiment_id}-r${realization_index:?}i${CMOR_INITIALIZATION_INDEX}p1f1_gr.nc
      else
        AREACELLA_CMOR_FILENAME=areacella_fx_${dataset}_${CMOR_EXP}_r${realization_index:?}i${CMOR_INITIALIZATION_INDEX}p1f1_gr.nc
      fi
EOF
      cat >> $script_name << EOF 
      CMOR_FX_DIR=\$RUN_dir/$cmor_outpath/\${CMOR_ACTIVITY_ID}/EC-Earth-Consortium/\${dataset}/\${CMOR_EXP}/r\${realization_index:?}i\${CMOR_INITIALIZATION_INDEX}p1f1/fx/areacella/gr/\$cmordatasetversion
EOF
      cat >> $script_name << "EOF" 
      mkdir -p "$CMOR_FX_DIR"
      [[ ! -f ${CMOR_FX_DIR}/$AREACELLA_CMOR_FILENAME ]] && cp ${PROJDIR}/fx_files/${IFS_resolution}/areacella_fx_EC-Earth3-CC_piControl_r0i0p0f0_gr.nc "${CMOR_FX_DIR}/$AREACELLA_CMOR_FILENAME"
    fi
  else
EOF
      cat >> $script_name << EOF 
  #if not first chunk, remove the files (because already generated at chunk 1
    find "\$RUN_dir/$cmor_outpath" -type d -name "*fx" -exec rm -rf {} +
  fi
  find "\$RUN_dir/$cmor_outpath" -type f -name "*Eyr*nc" -print0 | while IFS= read -r -d '' f; do
EOF
      cat >> $script_name << "EOF" 
      if ncdump -h "$f" | grep -q "frequency = \"yrPt" ; then
      year=$(basename "$f" | cut -f7 -d"_" | cut -c1-4)
      cdo "settaxis,${year}-12-31,23:00" "$f" "${f}2"
      cdo setreftime,1850-01-01,00:00 "${f}2" "$f"
      rm "${f}2"
    fi
  done
  # set the cmor process to complete
  touch "${PATH_FLAGS}/cmor_${component}_${CHUNK}"
}

#
#####################################################################################################################
# Globals: RUN_dir, CHUNK, cmor_dir, component
# Arguments:
#   None
# Returns:
#   None
# Purpose: Function to clean the cmorisation output by looking for .copy files and files with corrupted names.
#            When running cmorisation twice in the same CHUNK,  some files might end up duplicated or with a bad name.
#            This funciton cleans the cmor_${CHUNK} before it's transferred.
#
#####################################################################################################################
function clean_cmor() {
EOF
      cat >> $script_name << EOF 
  cmor_dir=\$RUN_dir/$cmor_outpath
EOF
      cat >> $script_name << "EOF" 
  cd "$cmor_dir"
  if [[ ${component} == "lpjg" ]]; then
    find "$cmor_dir" -type f -name "*nc" -print0 | while IFS= read -r -d '' f; do
      ncdump -h "$f" | grep -q "time:units = \"days since 1850-01-01\" ;" && ncatted -O -a units,time,m,c,"days since 1850-01-01 00:00:00" "$f"
    done
  fi
  find . -name "*.copy*" -type f -delete
}

#
#####################################################################################################################
# Globals: RUN_dir, CHUNK, Chunk_start_date, Chunk_end_date, cmor_dir, component
# Arguments:
#   None
# Returns:
#   None
# Purpose: Function to check number of timesteps of 6hourly files. This is a quick check previous to NCTIME which is run after transfer to esarchive
#
#####################################################################################################################
function check_cmor_timestep() {
EOF
      cat >> $script_name << EOF 
  cmor_dir=\$RUN_dir/$cmor_outpath
EOF
      cat >> $script_name << "EOF" 
  cd "$cmor_dir"
  #Check one directory of each frequency
  for freq in 6hr day mon; do
    dir_to_check=""
    for dir_to_check in $(find "$cmor_dir" -mindepth 9 -type d | grep -E "$freq" | head -1); do
      if [[ -n $dir_to_check ]]; then
        if [[ "$dir_to_check" == *6hr* ]]; then
          expected_timesteps=$((($(date -u -d $Chunk_end_date +%s) - $(date -u -d $Chunk_start_date +%s)) / 21600 + 4))
        elif [[ "$dir_to_check" == *day* ]]; then
          expected_timesteps=$((($(date -u -d $Chunk_end_date +%s) - $(date -u -d $Chunk_start_date +%s)) / 86400 + 1))
        else
          expected_timesteps=$((CHUNKSIZE))
        fi
        cd "$dir_to_check"
        number_of_timesteps=$(
          for f in *; do
            cdo -s ntime "$f"
          done | awk '{sum += $1} END {print sum}'
        )
        if [[ $number_of_timesteps != "$expected_timesteps" ]]; then
          errmsg="Found $number_of_timesteps timesteps in $dir_to_check, when we were expecting $expected_timesteps. Something might have gone wrong with CMORization. Chec\
k the outputs."
          # special exception for LPJG daily output in a leap year
          if [[ ${component} == "lpjg" ]] && [[ ${freq} == "day" ]] ; then
            if [[ ${number_of_timesteps} = 365 ]] && [[ ${expected_timesteps} = 366 ]] ; then
              echo "special exception for LPJG daily variable in $dir_to_check"
            elif [[ ${number_of_timesteps} = 366 ]] && [[ ${expected_timesteps} = 365 ]] ; then
              echo "special exception for LPJG daily variable in $dir_to_check"
            else
              echo "$errmsg"
              exit 1
            fi
          else
            echo "$errmsg"
            exit 1
          fi
        fi
        #echo $freq $dir_to_check $expected_timesteps $number_of_timesteps
      fi
    done
  done
}
EOF
      cat >> $script_name << EOF 
######################################################
#
# MAIN
#
######################################################

if [[ \${CMORIZATION} == '' ]] || [[ \${CMORIZATION} == 'TRUE' ]]; then
  # manu:set the cmorization component
  component=${component}  
  # prepare environment
  setenvcmor
  # execute cmorization for the related component
  cmor
  # clean garbage
  clean_cmor
  #check number of timesteps of cmor 6hourly output
  check_cmor_timestep
else
  echo \"CMORIZATION is disabled but the job \${TASKTYPE} is still in the job list of the experiment, please remove it if you are not using it\"
fi
EOF
    # component loop
    done
  # chunk loop
  done
# member loop
done 

# create the Parsl workflow file
launch_parsl_script_name=$LOGDIR/%EXPID%_%SDATE%_%MEMBER%_%CHUNK%_LAUNCH_PARSL.cmd

cat > $launch_parsl_script_name << EOF
#!/usr/bin/env python3
# 
# PARSL workflow  
# 
# Copyright (C) 2024 Manuel G. Marciani
# Earth Sciences Department - BSC-CNS 

import parsl
# for printing to stderr
import sys 
from parsl.config import Config
from parsl.providers import LocalProvider
from parsl.launchers import SrunLauncher
from parsl.channels import LocalChannel
from parsl.executors import HighThroughputExecutor
from parsl.addresses import address_by_interface
from parsl.app.app import bash_app

config = Config(
    executors=[
        HighThroughputExecutor(
            label='PM_HTEX_headless',
            provider=LocalProvider(
                nodes_per_block=%NODES%, 
                launcher=SrunLauncher(overrides='-c 56'),
                cmd_timeout=120,
                init_blocks=1,
                max_blocks=1
            ),
        )
    ],
    strategy=None,
)

parsl.load(config)
EOF

# loop over every member to be executed
for THIS_MEMBER_NUMBER in $(seq $number_of_members); do

  THIS_MEMBER=$(printf "%02d\n" $THIS_MEMBER_NUMBER)

  # loop over every chunk to be executed
  for THIS_CHUNK_NUMBER in $(seq $number_of_chunks); do 

    # define each of the component's 
    for component in %CMOR_WRAPPED_COMPONENTS%; do

      THIS_CHUNK=$(printf "%03d\n" $THIS_CHUNK_NUMBER)

      # here we do not use HEREDOCS because we want that bash evaluates the variables

      # NOTICE THAT IT DOES NOT HAVE THE FILE TYPE (e.g. whatever comes after the .)
      script_name=%EXPID%_%SDATE%_${THIS_MEMBER}_${THIS_CHUNK}_CMOR$(echo $component | tr '[:lower:]' '[:upper:]') 
        
      echo "@bash_app
def cmor${component}_${THIS_MEMBER}_${THIS_CHUNK}(dependency=None, stdout=\"$script_name.out\", stderr=\"$script_name.err\"):
    if dependency:
        # we wait for the dependency if we have it
        dependency.done()
    return f'echo \"${script_name} has dependency {dependency}.\"; /usr/bin/bash ${script_name}.cmd'
" >> $launch_parsl_script_name
    done
  done
done

# loop over every member to be executed
for THIS_MEMBER_NUMBER in $(seq $number_of_members); do

  THIS_MEMBER=$(printf "%02d\n" $THIS_MEMBER_NUMBER)

  # instruct to th run each of the components 
  for component in %CMOR_WRAPPED_COMPONENTS%; do 

    # loop over every chunk to be executed
    for THIS_CHUNK_NUMBER in $(seq $number_of_chunks); do
      THIS_CHUNK=$(printf "%03d\n" $THIS_CHUNK_NUMBER)
      PREVIOUS_CHUNK=$(printf "%03d\n" $(($THIS_CHUNK_NUMBER-1)))

      # output to the future object named like atm_01_002. This allows us to inspect the output afterwards
      if [[ "$THIS_CHUNK_NUMBER" == "1" ]]; then
        echo "${component}_${THIS_MEMBER}_${THIS_CHUNK} = cmor${component}_${THIS_MEMBER}_${THIS_CHUNK}()" >> $launch_parsl_script_name 
      else 
        echo "${component}_${THIS_MEMBER}_${THIS_CHUNK} = cmor${component}_${THIS_MEMBER}_${THIS_CHUNK}(${component}_${THIS_MEMBER}_${PREVIOUS_CHUNK})" >> $launch_parsl_script_name
      fi

    done
  done
done

# finally, set up and launch the in-situ workflow manager 

source %PARSL_VENV%/bin/activate

python3 $launch_parsl_script_name

